import json
from time import strftime, localtime


def save_file(file: bytes, path: str, name: str):
    f = open(path + "\\" + name, "wb")
    f.write(file)
    f.close()


def save_json(content, path: str, name: str):
    f = open(path + "\\" + name, "w", encoding="utf-8")
    f.write(json.dumps(content, ensure_ascii=False))
    f.close()


def get_time() -> str:
    return strftime("%Y-%m-%d %H:%M:%S", localtime())



from typing import Dict
from pbrm import update
from pbrm import config
from pbrm import delete
from pbrm import change_repository
from pbrm import pbrm_statistics
from pbrm import gif


def command_process(args: Dict, script_path: str, work_path: str):
    if args["update"]:
        update.update(config.cookie, config.repository_path, args["--skip-download"], args["--skip-meta"]
                      , args["--force-update"], args["--force-update-illust"], args["--force-update-meta"]
                      , args["--auto-remove"], config.ugoira == "gif")

    elif args["delete"]:
        delete.delete_illust(args["<PID>"], config.repository_path)
        print("deleted")

    elif args["config"]:
        if args["<CONFIG>"] == "cookie":
            if args["<CONTENT>"] is None:
                print(config.cookie)
            else:
                config.change_cookie(args["<CONTENT>"], script_path)
                print("changed")
        elif args["<CONFIG>"] == "ugoira":
            if args["<CONTENT>"] is None:
                print(config.ugoira)
            elif args["<CONTENT>"] == "raw" or args["<CONTENT>"] == "gif":
                if args["<CONTENT>"] == "gif":
                    gif.auto_transform(config.repository_path)
                config.change_ugoira(args["<CONTENT>"], script_path)
                print("changed")
            else:
                print("unknown config content")
        else:
            print("unknown config")

    elif args["set"]:
        if args["<PATH>"] is None:
            print(config.repository_path)
        elif args["<PATH>"] == "this":
            change_repository.change_repository(config.repository_path, work_path)
            config.change_repository(work_path, script_path)
        else:
            change_repository.change_repository(config.repository_path, args["<PATH>"])
            config.change_repository(args["<PATH>"], script_path)

    elif args["show"]:
        if args["unavailable"]:
            if args["--only-number"]:
                print(len(pbrm_statistics.statistics_unavailable(config.repository_path)))
            else:
                for i in pbrm_statistics.statistics_unavailable(config.repository_path):
                    print(i)
        elif args["unavailableSaved"]:
            if args["--only-number"]:
                print(len(pbrm_statistics.statistics_unavailableSaved(config.repository_path)))
            else:
                for i in pbrm_statistics.statistics_unavailableSaved(config.repository_path):
                    print(i)
        elif args["saved"]:
            if args["--only-number"]:
                print(len(pbrm_statistics.statistics_saved(config.repository_path)))
            else:
                for i in pbrm_statistics.statistics_saved(config.repository_path):
                    print(i)



import shutil
from pbrm import error
import os


def delete(path: str):
    shutil.rmtree(path)


def delete_illust(pid: str, path: str):
    illusts = [f for f in os.listdir(path) if os.path.isdir(os.path.join(path, f))]
    if pid not in illusts:
        raise error.IllustNotExists("pid " + pid + " is not existed")

    delete(path + "/" + pid)



import requests
import json
import pbrm


# 单页作品数量,可以决定获取收藏时连接次数(设太高谁知道会发生生么)
one_page_count = 48


# 获取收藏列表返回字典,会抛出GetBookmarksError
def get_bookmarks(cookie: str, user: str):
    header = {
        "cookie": cookie,
        "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) C"
                      "hrome/111.0.0.0 Safari/537.36",
        "referer": "https://www.pixiv.net/users/" + user + "/bookmarks/artworks",
        "sec-fetch-dest": "empty",
        "sec-ch-ua-platform": "\"Windows\"",
        "sec-ch-ua": "\"Google Chrome\";v=\"111\", \"Not(A:Brand\";v=\"8\", \"Chromium\";v=\"111\"",
        "pragma": "no-cache",
        "cache-control": "no-cache",
        "accept-encoding": "gzip, deflate, br",
        "accept": "application/json",
        "x-user-id": user
    }
    total = 0
    page = 1
    bookmarks = {"total": 0, "illust": []}

    # 分页循环获取所有收藏作品
    while 1:
        resp = requests.get("https://www.pixiv.net/ajax/user/{}/illusts/bookmarks?tag=&offset={}&limit={}&rest=show"
                            .format(user, str((page - 1) * one_page_count), str(one_page_count)), headers=header)\
                            .content.decode("utf-8")
        json_data = json.loads(resp)
        if json_data["error"]:
            raise pbrm.GetBookmarksError(json_data["message"])
        if total == 0:
            total = json_data["body"]["total"]
        bookmarks["illust"] += json_data["body"]["works"]
        if page * one_page_count >= total:
            break
        page += 1

    bookmarks["total"] = total
    return bookmarks



import requests
import json
import pbrm
from lxml import etree


# 验证cookie可用性,可用会返回一个包含账号id和账号名的字典,否则抛出CookieVerifyError
def cookie_verify(cookie: str):
    header = {
        "cookie": cookie,
        "referer": "https://www.pixiv.net/",
        "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) "
                      "Chrome/111.0.0.0 Safari/537.36",
        "sec-fetch-dest": "document",
        "sec-ch-ua-platform": "\"Windows\"",
        "sec-ch-ua": "\"Google Chrome\";v=\"111\", \"Not(A:Brand\";v=\"8\", \"Chromium\";v=\"111\"",
        "pragma": "no-cache",
        "cache-control": "no-cache",
        "accept-encoding": "gzip, deflate, br",
        "accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,"
                  "application/signed-exchange;v=b3;q=0.7"
    }
    http = requests.get("https://www.pixiv.net/", headers=header).content.decode("utf-8")
    content = etree.HTML(http).xpath("//meta[@name=\"global-data\"]/@content")
    user_data = json.loads(content[0])["userData"]
    if user_data is None:
        raise pbrm.CookieVerifyError("cookie无效,需要更换cookie")
    return {"userId": user_data["id"], "userName": user_data["pixivId"]}



import os
import zipfile
import subprocess

from pbrm import spider
from pbrm import utils
from pbrm import error


def save_log(log, path: str):
    utils.save_json(log, path, "log.json")


def save_meta(meta, path: str):
    utils.save_json(meta, path, "meta.json")


def save_img(meta, path: str, cookie: str):
    log = {"available": True, "updateTime": utils.get_time(), "illustType": 0, "pid": meta["id"]}
    save_log(log, path)
    pages = spider.get_pages(meta["id"], cookie)

    # 保存所有图片
    for page in pages:
        url = page["urls"]["original"]
        file = spider.image_download(url)
        file_name = url.split("/")[-1]
        utils.save_file(file, path, file_name)


def save_manga(meta, path: str, cookie: str):
    save_img(meta, path, cookie)


def save_ugoira(meta, path: str, cookie: str, save_gif: bool):
    log = {"available": True, "updateTime": utils.get_time(), "illustType": 2, "pid": meta["id"]}
    save_log(log, path)
    if not os.path.exists(path + "/images"):
        os.mkdir(path + "/images")
    ugoira_meta = spider.get_ugoira_meta(meta["id"], cookie)
    utils.save_json(ugoira_meta, path, "ugoira.json")
    ugoira = spider.ugoira_download(ugoira_meta["originalSrc"])
    utils.save_file(ugoira, path + "/images", "ugoira.zip")
    with zipfile.ZipFile(path + "/images/ugoira.zip", "r") as zip_file:  # 解压压缩包
        zip_file.extractall(path + "/images")
    os.remove(path + "/images/ugoira.zip")

    # 转gif
    if save_gif:
        frame_rate = 1000 / ugoira_meta["frames"][0]["delay"]
        file_name = ugoira_meta["frames"][0]["file"]
        extension = file_name.split(".")[1]
        length = len(file_name.split(".")[0])

        # ffmpeg合成gif
        result = subprocess.run("ffmpeg -framerate {} -i {}/%0{}d.{} -vf"
                                " \"split[s0][s1];[s0]palettegen[p];[s1][p]paletteuse\" {}.gif"
                                .format(str(frame_rate), path + "/images", str(length),
                                        extension, path + "/" + meta["id"])
                                , stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        if result.returncode != 0:
            raise error.CommandRunError("stdout: " + result.stdout.decode()
                                        + "\n--------------------------\nstderr: " + result.stderr.decode())


def save_illust(pid: str, path: str, cookie: str, save_gif: bool, skip_download: bool, skip_meta: bool):
    meta = spider.get_illust_meta(pid, cookie)
    if not skip_download:
        if meta["illustType"] == 0:
            save_img(meta, path, cookie)
        elif meta["illustType"] == 1:
            save_manga(meta, path, cookie)
        elif meta["illustType"] == 2:
            save_ugoira(meta, path, cookie, save_gif)
        else:
            raise error.UnSupportIllustType("UnSupport type: " + str(meta["illustType"]))
    if not skip_meta:
        save_meta(meta, path)


# 无效作品get_meta时会抛出错误,必须单独适配
def save_unavailable(meta, path):
    save_meta(meta, path)
    log = {"available": False, "updateTime": utils.get_time(), "illustType": None, "pid": meta["id"]}
    save_log(log, path)
    save_meta(meta, path)




import docopt
from pbrm import command_process
from pbrm import config
import os

doc = """Pixiv Bookmarks Repository Manager

Usage:
    pbrm update [-dmafin]
    pbrm delete <PID>
    pbrm config <CONFIG> [<CONTENT>]
    pbrm set [<PATH>]
    pbrm show (unavailable | unavailableSaved | saved) [-o]

Options:
    -d, --skip-download             跳过下载作品(仍会下载meta)
    -m, --skip-meta                 跳过下载meta数据
    -a, --auto-remove               更新同时删除已取消收藏的作品
    -f, --force-update              强制更新所有收藏(已保存的收藏将会重新下载，会跳过已失效的收藏)
    -i, --force-update-illust       强制更新作品本体
    -n, --force-update-meta         强制更新所有meta数据
    -o, --only-number               只显示作品数量而不是显示作品id
    
    <PID>                           作品id
    <CONFIG>                        配置项,可配置项为cookie,ugoira(动图保存方式,对应CONTENT有raw和gif,分别是多张图片保存和需要ffmpeg的gif保存)
    <CONTENT>                       配置项对应内容,为空时是查看配置
    <PATH>                          仓库位置路径,为空时为显示仓库路径,为this是使用当前路径,其余为指定路径
    
    update                          更新收藏
    delete                          删除指定收藏(仅支持pid指定)
    config                          查看或修改配置
    set                             设置仓库路径
    show                            显示统计信息
    unavailable                     失效且未保存的作品
    unavailableSaved                失效但已保存的作品
    saved                           已保存的作品

"""


def main():
    args = docopt.docopt(doc, version="Pixiv Bookmarks Repository Manager 0.0.1")
    work_path = os.getcwd().replace("\\", "/")
    script_path = os.path.dirname(__file__).replace("\\", "/")
    config.load(script_path + "/config.json")
    command_process.command_process(args, script_path, work_path)


if __name__ == "__main__":
    main()



import json
import subprocess
import os
from pbrm import error


def img_to_gif(match_path: str, save_path: str, frame_rate: float):
    result = subprocess.run("ffmpeg -framerate {} -i {} -vf"
                            " \"split[s0][s1];[s0]palettegen[p];[s1][p]paletteuse\" {}"
                            .format(str(frame_rate), match_path, save_path)
                            , stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    if result.returncode != 0:
        raise error.CommandRunError("stdout: " + result.stdout.decode()
                                    + "\n--------------------------\nstderr: " + result.stderr.decode())


def gif_illust(path: str):
    try:
        with open(os.path.join(path, "ugoira.json"), "r", encoding="utf-8") as f:
            ugoira_meta = json.loads(f.read())
    except FileNotFoundError:
        return

    with open(os.path.join(path, "log.json"), "r", encoding="utf-8") as f:
        name = json.loads(f.read())["pid"]

    if os.path.exists(os.path.join(path, name + ".gif")):
        return

    frame_rate = 1000 / ugoira_meta["frames"][0]["delay"]
    file_name = ugoira_meta["frames"][0]["file"]
    extension = file_name.split(".")[1]
    length = len(file_name.split(".")[0])

    img_to_gif(path + "/images/%0{}d.{}".format(str(length), extension), os.path.join(path, name + ".gif"), frame_rate)


def auto_transform(path: str):
    illusts = [f for f in os.listdir(path) if os.path.isdir(os.path.join(path, f))]
    for i in illusts:
        gif_illust(os.path.join(path, i))



import requests


header = {
    "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) C"
                  "hrome/111.0.0.0 Safari/537.36",
    "referer": "https://www.pixiv.net/",
    "sec-fetch-dest": "image",
    "sec-ch-ua-platform": "\"Windows\"",
    "sec-ch-ua": "\"Google Chrome\";v=\"111\", \"Not(A:Brand\";v=\"8\", \"Chromium\";v=\"111\"",
    "pragma": "no-cache",
    "cache-control": "no-cache",
    "accept-encoding": "gzip, deflate, br",
    "accept": "image/avif,image/webp,image/apng,image/svg+xml,image/*,*/*;q=0.8"
}


# 图片的bytes,需自行根据链接判断图片格式
def image_download(url: str) -> bytes:
    # 不管怎么样把看起来比较有用的都塞进去~
    h = header

    return requests.get(url, headers=h).content


# 按理返回的是一个压缩包,需要后续处理
def ugoira_download(url: str) -> bytes:

    h = header
    h["origin"] = "https://www.pixiv.net"
    h["sec-fetch-dest"] = "empty"
    h["accept"] = "*/*"

    return requests.get(url, headers=h).content



from .get_meta import get_illust_meta, get_ugoira_meta, get_pages
from .illust_download import ugoira_download, image_download
from .cookie_verify import cookie_verify
from .get_bookmarks import get_bookmarks



import os
import json


def statistics_unavailable(path: str):
    s = []
    with open(os.path.join(path, "bookmarks.json"), "r", encoding="utf-8") as f:
        bookmarks = json.loads(f.read())
    for i in bookmarks["illust"]:
        if i["userId"] == 0:
            with open(path + "/" + i["id"] + "/meta.json", "r", encoding="utf-8") as f:
                if json.loads(f.read())["userId"] == 0:
                    s.append(i["id"])
    return s


def statistics_unavailableSaved(path: str):
    s = []
    with open(os.path.join(path, "bookmarks.json"), "r", encoding="utf-8") as f:
        bookmarks = json.loads(f.read())
    for i in bookmarks["illust"]:
        if i["userId"] == 0:
            with open(path + "/" + i["id"] + "/meta.json", "r", encoding="utf-8") as f:
                if json.loads(f.read())["userId"] != 0:
                    s.append(i["id"])
    return s


def statistics_saved(path: str):
    illusts = [f for f in os.listdir(path) if os.path.isdir(os.path.join(path, f))]
    return illusts




import json
import os


cookie: str

ugoira: str

ffmpeg: str

repository_path: str


def load(path: str):
    global cookie
    global ugoira
    global ffmpeg
    global repository_path
    if not os.path.exists(path):
        config = {"cookie": "", "ugoira": "raw", "ffmpeg": "", "repositoryPath": ""}
        f = open(path, "w", encoding="utf-8")
        f.write(json.dumps(config, ensure_ascii=False))
        f.close()
        cookie = ""
        ugoira = "img"
        ffmpeg = ""
        repository_path = ""
    else:
        f = open(path, "r", encoding="utf-8")
        j = f.read()
        f.close()
        js = json.loads(j)
        cookie = js["cookie"]
        ugoira = js["ugoira"]
        ffmpeg = js["ffmpeg"]
        repository_path = js["repositoryPath"]


def change_cookie(c: str, path):
    config = {"cookie": c, "ugoira": ugoira, "ffmpeg": ffmpeg, "repositoryPath": repository_path}
    with open(os.path.join(path, "config.json"), "w", encoding="utf-8") as f:
        f.write(json.dumps(config))


def change_ugoira(u: str, path):
    config = {"cookie": cookie, "ugoira": u, "ffmpeg": ffmpeg, "repositoryPath": repository_path}
    with open(os.path.join(path, "config.json"), "w", encoding="utf-8") as f:
        f.write(json.dumps(config))


def change_repository(p: str, path):
    config = {"cookie": cookie, "ugoira": ugoira, "ffmpeg": ffmpeg, "repositoryPath": p}
    with open(os.path.join(path, "config.json"), "w", encoding="utf-8") as f:
        f.write(json.dumps(config))



import json

from pbrm import spider
from pbrm import save_illust
from pbrm import utils
import os
from pbrm import delete


def work_exists(pid: str, path: str, unavailable: bool):
    if os.path.exists(path + "/" + pid) and os.path.isdir(path + "/" + pid):
        if len(os.listdir(path + "/" + pid)) > (1 if unavailable else 2):
            return True
        else:
            return False

    return False


# 失效的作品的id和userId是整形而不是字符串
def update(cookie: str, path: str, skip_download: bool, skip_meta: bool, force_update: bool, force_update_illust: bool
           , force_update_meta: bool, auto_remove: bool, save_gif: bool):

    if not (os.path.exists(path) and os.path.isdir(path)):
        os.mkdir(path)

    # 合并force_update_illust和force_update_meta
    if force_update_meta is True and force_update_illust is True:
        force_update = True
    user = spider.cookie_verify(cookie)
    bookmarks = spider.get_bookmarks(cookie, user["userId"])
    log = {"updateTime": utils.get_time(), "total": bookmarks["total"], "updated": 0, "unavailable": 0}
    print("userId: {} userName: {} total: {}".format(user["userId"], user["userName"], bookmarks["total"]))

    i = 0
    all_illust = []
    for illust in bookmarks["illust"]:
        i = i + 1
        all_illust.append(str(illust["id"]))
        print("update: " + str(illust["id"]) + " process: {}/{}".format(str(i), bookmarks["total"]))
        if not force_update:
            if work_exists(str(illust["id"]), path, illust["userId"] == 0):
                if force_update_meta != force_update_illust and illust["userId"] != 0:
                    # 将force当skip用
                    log["updated"] += 1
                    save_illust.save_illust(illust["id"], path + "/" + illust["id"], cookie, save_gif
                                            , not force_update_illust, not force_update_meta)
                continue

        if not (os.path.exists(path + "/" + str(illust["id"])) and os.path.isdir(path + "/" + str(illust["id"]))):
            os.mkdir(path + "/" + str(illust["id"]))

        if illust["userId"] == 0:
            log["unavailable"] += 1
            save_illust.save_unavailable(illust, path + "/" + str(illust["id"]))
            continue

        log["updated"] += 1
        save_illust.save_illust(illust["id"], path + "/" + illust["id"], cookie, save_gif, skip_download, skip_meta)

    if auto_remove:
        illusts = [f for f in os.listdir(path) if os.path.isdir(os.path.join(path, f))]
        for i in illusts:
            if i not in all_illust:
                delete.delete(os.path.join(path, i))

    with open(os.path.join(path, "log.json"), "w", encoding="utf-8") as f:
        f.write(json.dumps(log))

    with open(os.path.join(path, "bookmarks.json"), "w", encoding="utf-8") as f:
        f.write(json.dumps(bookmarks))




# 获取元数据错误


class PixivGetError(Exception):
    message: str

    def __init__(self, message: str):
        self.message = message

    def __str__(self):
        return "Error Message: " + self.message


class GetMetaError(PixivGetError):
    pid: str  # 作品id

    def __init__(self, pid: str, message: str):
        self.pid = pid
        self.message = message

    def __str__(self):
        return "illustId: " + self.pid + " message: " + self.message


class GetIllustMetaError(GetMetaError):
    pass


class GetUgoiraMetaError(GetMetaError):
    pass


class GetPagesError(GetMetaError):
    pass


class GetBookmarksError(PixivGetError):
    pass


class CookieVerifyError(PixivGetError):
    pass


class UnSupportIllustType(PixivGetError):
    pass


class CommandRunError(PixivGetError):
    pass


class IllustNotExists(PixivGetError):
    pass



import requests
import json
import pbrm


# 通用header
header = {
    "cookie": "",
    "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) "
                  "Chrome/111.0.0.0 Safari/537.36",
    "sec-ch-ua-platform": "\"Windows\"",
    "sec-ch-ua": "\"Google Chrome\";v=\"111\", \"Not(A:Brand\";v=\"8\", \"Chromium\";v=\"111\"",
    "referer": "",
    "pragma": "no-cache",
    "cache-control": "no-cache",
    "accept-encoding": "gzip, deflate, br",
    "accept": "*/*"
}

referer = "https://www.pixiv.net/artworks/{}"


# 获取动图数据
def get_ugoira_meta(pid: str, cookie: str):
    h = header
    h["cookie"] = cookie
    h["referer"] = referer.format(pid)

    meta_data = requests.get("https://www.pixiv.net/ajax/illust/{}/ugoira_meta".format(pid), headers=h) \
        .content.decode("utf-8")
    json_data = json.loads(meta_data)
    if json_data["error"]:
        raise pbrm.GetUgoiraMetaError(pid, json_data["message"])     # 抛出error

    return json_data["body"]


# 获取作品数据
def get_illust_meta(pid: str, cookie: str):
    h = header
    h["cookie"] = cookie
    h["referer"] = referer.format(pid)

    meta_data = requests.get("https://www.pixiv.net/ajax/illust/" + pid, headers=h).content.decode("utf-8")
    json_data = json.loads(meta_data)
    if json_data["error"]:
        raise pbrm.GetIllustMetaError(pid, json_data["message"])

    return json_data["body"]


def get_pages(pid: str, cookie: str):
    h = header
    h["referer"] = referer.format(pid)
    h["cookie"] = cookie
    pages_data = requests.get("https://www.pixiv.net/ajax/illust/{}/pages".format(pid), headers=h).content\
        .decode("utf-8")
    pages_json = json.loads(pages_data)
    if pages_json["error"]:
        raise pbrm.GetPagesError(pid, pages_json["message"])

    return pages_json["body"]



from setuptools import setup, find_packages

setup(
    name='pbrm',
    version='0.0.3',
    description='A pixiv bookmarks local backup tool',
    author='stay_miku',
    author_email='root@staymiku.top',
    url='https://github.com/yourusername/myproject',
    packages=find_packages(),
    install_requires=[
        "requests",
        "lxml",
        "docopt"
    ],
    entry_points={
        'console_scripts': [
            'pbrm=pbrm.main:main',
        ],
    }
)
